/*
*************************************************************************
*                                                                       *
* sfcIndex.C                                                            *
*                                                                       *
* Class sfcIndex defines an index generated by the chosen SFC mapping   *
* i.e. specified space dimensions, base used and mapping function.      *
* sfcIndex is an ordering vector of sfcIndexUnits; each sfcIndexUnits   *
* defined as Dim Base-b digts, and defining the index at each level of  *
* the sfc. The most-significant sfcIndexUnit corresponds to level 1     *
* Class sfcIndex is derived from BitVec.				*
*                                                                       *
* Author:  Manish Parashar (parashar@cs.utexas.edu)                     *
*                                                                       *
*************************************************************************
*/

#include "sfcIndex.h"

#ifndef __ENABLE_INLINE__
#include "sfcIndex.inline"
#endif

/*
*************************************************************************
*                                                                       *
* unsigned sfcIndex::sfcGetIndex() const				*
*                                                                       *
* Member function sfcGetIndex returns the (unsigned) integer value of 	*
* the index.								*
*									*
* Author:  Manish Parashar (parashar@cs.utexas.edu)                     *
*                                                                       *
*************************************************************************
*/

unsigned sfcIndex::sfcGetIndex() const
  {
   unsigned digit = 0;
   if(sfcbase == 2)
     {
      for (int i=1;i<=sfclevs;i++)
        digit = (digit<<sfcdim) | sfcGetDigit(i);
     }
   else
     {
      register int off = 0;
      for (int i=1;i<=sfclevs;i++) {
        off = (i-1)*sfcdigitw;
        for (int j=0;j<sfcdim;j++)
         digit = digit*sfcbase + IsolateBit(off+(j*sfcbasew),sfcbasew);
      }
     }
 
   return digit;
  }
 
/*
*************************************************************************
*                                                                       *
* unsigned sfcIndex::sfcGetIndex(const int lev) const			*
*                                                                       *
* Member function sfcGetIndex returns the (unsigned) integer value of 	*
* the index at level lev; i.e. only the first lev digits of the 	*
* sfcIndex.								*
*									*
* Author:  Manish Parashar (parashar@cs.utexas.edu)                     *
*                                                                       *
*************************************************************************
*/

unsigned sfcIndex::sfcGetIndex(const int lev) const
  {
   assert(lev > 0 && lev <= sfclevs);

   unsigned digit = 0;
   if(sfcbase == 2)
     {
      for (int i=1;i<=lev;i++)
        digit = (digit<<sfcdim) | sfcGetDigit(i);

      return (digit);

      /*  left shift by (sfclevs-lev)*sfcdigitw bits.
      //return (digit << (sfclevs-lev)*sfcdim);
      */
     }
   else
     {
      register int off = 0;
      int i,j;
      for (i=1;i<=lev;i++) {
        off = (i-1)*sfcdigitw;
        for (j=0;j<sfcdim;j++)
         digit = digit*sfcbase + IsolateBit(off+(j*sfcbasew),sfcbasew);
      }
      /*  left shift by (sfclevs-lev)*sfcdigitw bits.
      //for (i=lev+1;i<=sfclevs;i++) {
      //  for (j=0;j<sfcdim;j++)
      //   digit = digit*sfcbase;
      //}
      */

      return digit;
     }
  }
 
/*
*************************************************************************
*                                                                       *
* unsigned sfcIndex::sfcGetIndex(const int lev, const int nlev) const	*
*                                                                       *
* Member function sfcGetIndex returns the (unsigned) integer value of 	*
* the index from level lev-nlev+1 to level lev; 			*
*									*
* Author:  Manish Parashar (parashar@cs.utexas.edu)                     *
*                                                                       *
*************************************************************************
*/

unsigned sfcIndex::sfcGetIndex(const int lev, const int nlev) const
  {
   assert(lev > 0 && lev <= sfclevs);

   unsigned digit = 0;
   if(sfcbase == 2)
     {
      for (int i=lev-nlev+1;i<=lev;i++)
        digit = (digit<<sfcdim) | sfcGetDigit(i);

      return digit;
     }
   else
     {
      register int off = 0;
      int i,j;
      for (i=lev-nlev+1;i<=lev;i++) {
        off = (i-1)*sfcdigitw;
        for (j=0;j<sfcdim;j++)
         digit = digit*sfcbase + IsolateBit(off+(j*sfcbasew),sfcbasew);
      }

      return digit;
     }
  }
 
/*
*************************************************************************
*                                                                       *
* void sfcIndex::sfcSetIndex(const unsigned index)			*
*                                                                       *
* Member function sfcSetIndex set the (unsigned) integer value of 	*
* the sfcIndex to "index".						*
*									*
* Author:  Manish Parashar (parashar@cs.utexas.edu)                     *
*                                                                       *
*************************************************************************
*/

void sfcIndex::sfcSetIndex(const unsigned index)
  {
   register unsigned mask = ~((~0)<<sfcdigitw);

   for (int i=1;i<=sfclevs;i++)
     sfcSetDigit(i,(index>>sfcdigitw*(sfclevs-i))&mask);
  }
 
/*
*************************************************************************
*                                                                       *
* void sfcIndex::sfcSetIndex(const unsigned index, const int lev)	*
*                                                                       *
* Member function sfcSetIndex sets the (unsigned) integer value of 	*
* the sfcIndex at level lev (i.e. MSB lev*sfcdigitw bits) to "index".	*
*									*
* Author:  Manish Parashar (parashar@cs.utexas.edu)                     *
*                                                                       *
*************************************************************************
*/

void sfcIndex::sfcSetIndex(const unsigned index, const int lev)
  {
   assert(lev > 0 && lev <= sfclevs);

   register unsigned mask = ~((~0)<<sfcdigitw);

   for (int i=1;i<=lev;i++)
     sfcSetDigit(i,(index>>sfcdigitw*(lev-i))&mask);
  }
 
/*
*************************************************************************
*                                                                       *
* void sfcIndex::sfcSetIndex(const unsigned index, const int lev,	*
*						   const int nlev)	*
*                                                                       *
* Member function sfcSetIndex sets the (unsigned) integer value of 	*
* the sfcIndex from lev to lev+nlev-1 to "index".			*
*									*
* Author:  Manish Parashar (parashar@cs.utexas.edu)                     *
*                                                                       *
*************************************************************************
*/

void sfcIndex::sfcSetIndex(const unsigned index, const int lev, 
						 const int nlev)
  {
   assert(lev >= nlev && lev <=sfclevs);

   register unsigned mask = ~((~0)<<sfcdigitw);

   for (int i=nlev-1;i>=0;i--)
     sfcSetDigit(lev-i,(index>>(sfcdigitw*i))&mask);
  }
 
/*
*************************************************************************
*                                                                       *
* void sfcIndex::sfcMaxIndex(int const lev)				*
*                                                                       *
* Author:  Manish Parashar (parashar@cs.utexas.edu)                     *
*                                                                       *
*************************************************************************
*/

void sfcIndex::sfcMaxIndex(int const lev)
  {
   assert(lev > 0 && lev <= sfclevs);
   for (int i=1;i<=lev;i++) sfcMaxDigit(i);
  }
 
/*
*************************************************************************
*                                                                       *
* void sfcIndex::sfcCreateLocNum(unsigned const *coords)		*
*                                                                       *
* Member function sfcCreateLocNum given the coordinates as follows:	*
* For example (3-D) let the coordinates be:				*
*    x = xb1 xb2 xb3 ......						*
*    y = yb1 yb2 yb3 ......						*
*    z = zb1 zb2 zb3 ......						*
* where xbi, ybj, zbk are base sfcbase (sfcbasew bits).			*
* Then the location number is defined as follows:			*
*    (xb1 yb1 zb1) (xb2 yb2 zb2) (xb3 yb3 zb3)) .....			*
*									*
* Author:  Manish Parashar (parashar@cs.utexas.edu)                     *
*                                                                       *
*************************************************************************
*/

void sfcIndex::sfcCreateLocNum(unsigned const *coords)
  {
   ResetVec();
   for (int i=sfclevs;i>0;i--)
     for (int j=0;j<sfcdim;j++) 
       for (int k=0;k<sfcbasew;k++) 
         if (coords[j]>>((sfclevs-i)*sfcbasew+(sfcbasew-k-1)) & (unsigned)1)
           sfcSetDigitBit(i,j*sfcbasew+k);
  }

/*
*************************************************************************
*                                                                       *
* void sfcIndex::sfcGetCoords(unsigned *coords)				*
*                                                                       *
* Member function sfcGetCoords returns an array of coordinates from the *
* locations number.							*
* For example (3-D) let the location number be:				*
*    (xb1 yb1 zb1) (xb2 yb2 zb2) (xb3 yb3 zb3)) .....			*
* where xbi, ybj, zbk are base sfcbase (sfcbasew bits).			*
* Then the cordinates are as follows:					*
*    x = xb1 xb2 xb3 ......						*
*    y = yb1 yb2 yb3 ......						*
*    z = zb1 zb2 zb3 ......						*
*									*
* Author:  Manish Parashar (parashar@cs.utexas.edu)                     *
*                                                                       *
*************************************************************************
*/

void sfcIndex::sfcGetCoords(unsigned *coords) const
  {
   int i;
   for (i=0;i<sfcdim;i++) 
       coords[i] = 0;

   register int off=0;
   for (i=1;i<=sfclevs;i++) {
     off = (i-1)*sfcdigitw;
     for (int j=0;j<sfcdim;j++)  {
       coords[j] = ((coords[j])<<sfcbasew) | IsolateBit(off+(j*sfcbasew),sfcbasew);
     }
   }
  }

ostream&  operator << (ostream& os, sfcIndex const &si)
  {
   if (&si == (sfcIndex *)NULL) return os;
   for (int i=1;i<=si.sfclevs;i++) os << si.sfcGetIndex(i,1);
   return os;
  }
